import unittest
from unittest.mock import MagicMock, patch

from freezegun import freeze_time
from jwt.exceptions import InvalidIssuerError

from ut_jwt.tokens import UTToken
from ut_jwt.exceptions import InvalidHeaderError
from .keys import PRIVATE_KEY, PUBLIC_KEY


class TestUTToken(unittest.TestCase):
    test_payload = {'iss': 'test_issuer', 'aud': 'test_audience', 'iat': 0, 'exp': 100}
    # Encoded version of test_payload with PRIVATE_KEY
    test_token = ("eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6InRlc3Rfa2lkIn0."
                  "eyJpc3MiOiJ0ZXN0X2lzc3VlciIsImF1ZCI6InRlc3RfYXVkaWVuY2UiLCJpYXQiOjAs"
                  "ImV4cCI6MTAwfQ."
                  "fWXOW_q8PVNlemG3O-zJypZwRR-JUhJEwhp2KUzy4GoLQy5t-l28Y15OB_LjAEhp7v2v"
                  "ApjKz58bY3NcAY6tQ6iouNpuOG7jks8E2dLa03lWpDf57i-52dT2xq9nt0vsZdcqPm17"
                  "Hvesu2M9h-AaTADBP8vgjGMo4pIpD2392BiCeg5zoWt2PIwiXPdxcdCKDaiJbWyJaNBS"
                  "xU3LUwMSnSI2BEExBU1YFer90Lr9v3-0bHb2fFDfJVSp2T-Ketslh_1v3Dk7uToQmHdk"
                  "kout9NOq1bSj1RbK3MgZD0SVobYpK2eFdlEKmaMTHL3MWaWIokPD4RFXTGfUY69U8FDU"
                  "HPvfPEdQwdSazxqpXJki-GD-mpJ3EWTb1u-SJ6PSaeT1LqYebQwXRZW37vJBh_E8e9Ss"
                  "yXKkmhVkQcewQ5N5bu1B2pRb72cjlOJz3-oiAVbh3Md0eR4hy1kUU7e7S4QGloZiIk0x"
                  "Gu-Y-ER21bec_Va5lUzd9pyU3qdxLyNXlOa9gpeN9tdHTekuvixxSILLjGr9ZBTyNNHR"
                  "AmZxm-_Na_j7Y8lFJUplOr_c7fopkmlUyWa9nMDGLHx0U4rtV0orcWIZvy1Cjm1HwzFB"
                  "fJf-2lckXz-e2aIOAk2JSMY6syuHnQ_lyJaF43lyS228y0OOZOHEKWW051cU03SWEzic"
                  "Sso")

    @patch("requests.get")
    def test_init_sets_fields_correctly(self, mock_requests_get):
        # Arrange
        token = self.test_token
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        path_to_public_keys = "/test/public/keys"
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act
        with freeze_time("1970-01-01"):
            ut_token = UTToken(
                token=token,
                accepted_issuers=accepted_issuers,
                expected_audience=expected_audience,
                path_to_public_keys=path_to_public_keys,
            )

            # Assert
            self.assertEqual(ut_token.token, token)
            self.assertEqual(ut_token.accepted_issuers, accepted_issuers)
            self.assertEqual(ut_token.expected_audience, expected_audience)
            self.assertEqual(ut_token.path_to_public_keys, path_to_public_keys)
            self.assertDictEqual(ut_token.payload, self.test_payload)

    @patch("requests.get")
    def test_init_decodes_token_correctly(self, mock_requests_get):
        # Arrange
        token = self.test_token
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        path_to_public_keys = "/test/public/keys"
        header = {"kid": "test_kid"}
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act
        with freeze_time("1970-01-01"):
            ut_token = UTToken(
                token=token,
                accepted_issuers=accepted_issuers,
                expected_audience=expected_audience,
                path_to_public_keys=path_to_public_keys
            )

        # Assert
        self.assertEqual(ut_token.payload, self.test_payload)

    @patch("requests.get")
    def test_init_raises_error_on_invalid_header(self, mock_requests_get):
        # Arrange
        no_kid_token = ("eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJ0ZXN0X2lzc3Vlc"
                        "iIsImF1ZCI6InRlc3RfYXVkaWVuY2UiLCJpYXQiOjAsImV4cCI6MTAwfQ.GdfW"
                        "TvMYYlJ93zCRGIKBhp_QTgd3XGtx-n7LRBtQVfsP7YKpuleYXSPzud-N8yhnm_"
                        "Z88qzSGJPNer9uvVU9qIDfiQtozXwvseIuQ9-aKx25rJphPgLrVhKpLrb5lONv"
                        "0QemTItly5VuM7KGp6IACR3erfNza1QVupCpvY9NE7QOl-8U9da75c7-8o8Kyi"
                        "w4aUwM4jNmBzzZNyJhrmUVQ64vga4PqHdJSQ1kgf6QK1i7t08-Povgk1NIWSSP"
                        "Ps6LrzZfINZZbhl8CObDZwBWSSOAUyF2q2fksvDl3DIXp6EB8A4VieGqh7G56u"
                        "OiX4Nq8p-XGrRuMVKh0FkPpZGFTufrn_KNfaWRBZm18hYR5rwc6l-_9YXuQlEt"
                        "dhqdZY2p33XEP_Pbc6zYejaoTLnASlzjHXtl2JIgQnQMby_3DFzMEV7O64zd3d"
                        "Gh6d203U6hZfVciVNLIjqQcpqsjGLRdzaghNoTg3W9X_zSwpOzk0mA8mLzujtr"
                        "82pO-7HbxsG_d7dcv-e7XvhAqMvavpctPv7zWWUkmNfg8i1l4-N31SuolEYtbG"
                        "psKbwy8-3pu5YTpQvpIPaapvGD6r18xyQv8JgbWlvgUsLc2IQWO47JUOP_RG5g"
                        "oPOWIkf7V9mjLwj8uRzogyh14XAzIdbXvCyK2kILA2D8uL83j6LIYWEFCJQ")
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        path_to_public_keys = "/test/public/keys"
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act & Assert
        with self.assertRaises(InvalidHeaderError) as cm:
            with freeze_time("1970-01-01"):
                UTToken(
                    token=no_kid_token,
                    accepted_issuers=accepted_issuers,
                    expected_audience=expected_audience,
                    path_to_public_keys=path_to_public_keys,
                    private_key=PRIVATE_KEY
                )
        thrown_exception = cm.exception
        self.assertEqual(str(thrown_exception),
                         'Token header is missing the "kid" field')

    def test_init_raises_error_on_empty_private_key(self):
        # Arrange
        empty_private_key = b""

        # Act & Assert
        with self.assertRaises(ValueError):
            UTToken(private_key=empty_private_key)

    def test_str_encodes_payload_correctly(self):
        with freeze_time("1970-01-01"):
            # Arrange
            token = UTToken(private_key=PRIVATE_KEY)
            token.create_payload("test_issuer", "test_audience")
            expected_str = ("eyJhbGciOiJSUzI1NiIsImtpZCI6InRlc3RfaXNzdWVyIiwidHlwIjoiSl"
                            "dUIn0.eyJpc3MiOiJ0ZXN0X2lzc3VlciIsImF1ZCI6InRlc3RfYXVkaWVu"
                            "Y2UiLCJpYXQiOjAsImV4cCI6MzAwfQ.gHZCNfLcqS12pitAycHNBKLWVH-"
                            "OmnslrONRXz2VzNAdmJQiFl1twauBXKDqAntetOyTifw_pmjFWc-CxhEB1"
                            "_uci0NsCY6yPs865iQLatQLO4NCTqnf3-I8YAtErhQRThZJ2Ykv6vrv6UG"
                            "xMzmFViWiHBphqhOuJvsu0nLA3ZlMInK9X2PEzw1bpI_e01ehazzKxcluy"
                            "sv32pIVNM4Ksu_20-jKzVIe1JlPTNcGqtu-5A04Wl1blVVI_MsEa6cI1me"
                            "YCx_H57cO7IQAk7qysAN5SVeFOaptpwOpSQtoUffljJbfNn_AOPUnmmMBA"
                            "Ixvu3A3cko_GFf29HFiDssx6NR5QK490G_NVaoFaazcC_v-o2zwb4tCp6X"
                            "fC4X8EBUX7d-pVqpdjnj3k9UWdihMXNQ71aFhuc5DccvnC2eLc7WeBzwpq"
                            "zvf-MRRND1jEI0eM15hS2n1qDwzLEMaObHovkpKIt97FEGSnj9rYo6YljM"
                            "iO8Tz4MgQ7FzO0aNVPS76j_EaOh7wv198-Mtby2EAfRzh9AtB6Q4TFu5Ar"
                            "18H7IltYDDZKBLPqQg9DaE7lXV5dzKI5VAOA-TRLYSPIl4yBaBqj8n0Xlu"
                            "dZ2WYdHW0MH7wqBBM1v8eL94EJK-O9PybJU_xUisvnrPVXZyie6ALQXLMu"
                            "dX3lE-Nspc4Y-gbwDg")

            # Act
            actual_str = str(token)

            # Assert
            self.assertEqual(actual_str, expected_str)

    def test_create_payload_sets_values_correctly(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)

        # Act
        token.create_payload("issuer", "audience")

        # Assert
        self.assertEqual(token["iss"], "issuer")
        self.assertEqual(token["aud"], "audience")

    def test_create_payload_raises_error_for_empty_iss(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)

        # Act/Assert
        with self.assertRaises(ValueError):
            token.create_payload("", "audience")

    def test_create_payload_raises_error_for_empty_aud(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)

        # Act/Assert
        with self.assertRaises(ValueError):
            token.create_payload("issuer", "")

    def test_create_payload_sets_iat_and_exp(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)

        # Act
        token.create_payload("issuer", "audience")

        # Assert
        self.assertIn("iat", token)
        self.assertIn("exp", token)

    def test_validate_iss_raises_error_for_invalid_iss(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)
        token.create_payload("issuer", "audience")
        token.accepted_issuers = ["other_issuer"]

        # Act/Assert
        with self.assertRaises(InvalidIssuerError):
            token.validate_iss()

    def test_validate_iss_does_not_raise_error_for_valid_iss(self):
        # Arrange
        token = UTToken(private_key=PRIVATE_KEY)
        token.create_payload("issuer", "audience")
        token.accepted_issuers = ["issuer"]

        # Act/Assert
        token.validate_iss()  # should not raise an error

    @patch("requests.get")
    def test_repr_returns_expected_payload(self, mock_requests_get):
        # Arrange
        token = self.test_token
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        path_to_public_keys = "/test/public/keys"
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act
        with freeze_time("1970-01-01"):
            ut_token = UTToken(
                token=token,
                accepted_issuers=accepted_issuers,
                expected_audience=expected_audience,
                path_to_public_keys=path_to_public_keys,
            )

            # Assert
            self.assertEqual(repr(ut_token), str(self.test_payload))

    @patch("requests.get")
    def test_delitem_removes_payload_item_correctly(self, mock_requests_get):
        # Arrange
        token = self.test_token
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        expected_new_payload = {'aud': 'test_audience', 'iat': 0, 'exp': 100}
        path_to_public_keys = "/test/public/keys"
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act
        with freeze_time("1970-01-01"):
            ut_token = UTToken(
                token=token,
                accepted_issuers=accepted_issuers,
                expected_audience=expected_audience,
                path_to_public_keys=path_to_public_keys,
            )

            del ut_token["iss"]
            # Assert
            self.assertDictEqual(ut_token.payload, expected_new_payload)

    @patch("requests.get")
    def test_get_returns_item_correctly(self, mock_requests_get):
        # Arrange
        token = self.test_token
        accepted_issuers = ["test_issuer"]
        expected_audience = ["test_audience"]
        path_to_public_keys = "/test/public/keys"
        mock_requests_get.return_value.raise_for_status = MagicMock()
        mock_requests_get.return_value.content = PUBLIC_KEY

        # Act
        with freeze_time("1970-01-01"):
            ut_token = UTToken(
                token=token,
                accepted_issuers=accepted_issuers,
                expected_audience=expected_audience,
                path_to_public_keys=path_to_public_keys,
            )

            # Assert
            self.assertEqual(ut_token.get("iss"), "test_issuer")
