from datetime import datetime, timedelta

import jwt
import requests

from .exceptions import InvalidHeaderError


class UTToken:
    def __init__(self, token=None, accepted_issuers=None, expected_audience=None,
                 path_to_public_keys="", private_key=b""):
        self.token = token
        self.accepted_issuers = accepted_issuers or []
        self.expected_audience = expected_audience or []
        self.path_to_public_keys = path_to_public_keys
        self.private_key = private_key

        if token:
            header = jwt.get_unverified_header(self.token)
            try:
                kid = header["kid"]
                public_key_uri = f"{self.path_to_public_keys}/{kid}.key.pub"
                resp = requests.get(public_key_uri)
                resp.raise_for_status()
                public_key = resp.content
                self.payload = jwt.decode(
                    self.token,
                    public_key,
                    algorithms=["RS256"],
                    audience=self.expected_audience,
                    options={"require": ["iss", "aud", "iat", "exp"]}
                )
                # Validate iss claim after decode, because we don't know it in advance
                self.validate_iss()
            except KeyError:
                raise InvalidHeaderError("kid")
        else:
            # Create a new token
            self.payload = {}
            if self.private_key == b"":
                raise ValueError('"private_key" is not allowed to be empty for new '
                                 'tokens')

    def __repr__(self):
        return repr(self.payload)

    def __getitem__(self, key):
        return self.payload[key]

    def __setitem__(self, key, value):
        self.payload[key] = value

    def __delitem__(self, key):
        del self.payload[key]

    def __contains__(self, key):
        return key in self.payload

    def get(self, key, default=None):
        return self.payload.get(key, default)

    def __str__(self):
        """Signs and returns a token as a base64 encoded string."""
        headers = {
            "alg": "RS256",
            "typ": "JWT",
            "kid": self.get("iss")
        }
        return jwt.encode(
            headers=headers,
            payload=self.payload,
            key=self.private_key,
            algorithm="RS256"
        )

    def validate_iss(self):
        if self.payload["iss"] not in self.accepted_issuers:
            raise jwt.InvalidIssuerError("Invalid issuer")

    def create_payload(self, iss: str, aud: str):
        if iss == "":
            raise ValueError('"iss" is not allowed to be empty')
        if aud == "":
            raise ValueError('"aud" is not allowed to be empty')
        now_ts = datetime.now()
        iat = int(now_ts.timestamp())
        exp = int((now_ts + timedelta(minutes=5)).timestamp())

        self.__setitem__("iss", iss)
        self.__setitem__("aud", aud)
        self.__setitem__("iat", iat)
        self.__setitem__("exp", exp)
